//==============================================================================
// Utilitaires pour rotations et quaternions
// Copyright (C) Michel Llibre.
//==============================================================================
// Dans les fonctions de ce fichiers les quaternions sont de simples
// vecteurs 4x1 avec la composante scalaire en premier.
// q = qconj(qa)        conjugué
// q = qinv(qa)         inverse
// q = qmul(qa,qb)      cad  q = qa*qb
// q = qrot(v3)         cad  q = vrot2qrot(v3)
// q = qrot(m33)        cad  q = mrot2qrot(m33)
// q = qrot(ang, iaxe)  cad  q = angix2qrot(ang, iaxe)
// q = qrot(ang, v)     cad  q = angvu2qrot(ang, v/||v||)
// q = qprox(qa,qb)     cad  q = (+/-)qa (le plus proche de qb)
// m = mrot(qrot)       cad  m = qrot2mrot(qrot)
// m = mrot(ang, iaxe)  cad  m = angix2mrot(ang, iaxe)
// v = vrot(m)          Vecteur rotation associe a la matrice unitaire m
// v = vrot(q)          Vecteur rotation associe au quaternion unitaire q
// [a,u] = vrot(m)      Angle de rotation et vect-unit de la matrice unitaire m
// [a,u] = vrot(q)   Angle de rotation et vect-unit du quaternion unitaire q
//  Angs = vrot(m,'uvw')  Angles d'Euler-Cardan de la matrice unitaire m
//           les lettres u, v et w specifient les axes successifs de
//           rotation. Elles doivent etre une des trois lettres x, y ou z.
// Angs = vrot(q,'uvw')  Angles d'Euler-Cardan extraits du quaternion unitaire q
//------------------------------------------------------------------------------

//-------------------------------------------------------------------
// qconj : Quaternion conjugué
//-------------------------------------------------------------------
function Qa = qconj(Qb)
    Qa = [Qb(1) ; -Qb(2) ; -Qb(3) ; -Qb(4)] ;
endfunction

//-------------------------------------------------------------------
// qinv : Quaternion inverse
//-------------------------------------------------------------------
function q = qinv(qa)
   q = qconj(qa)/(qa'*qa);
endfunction

//-------------------------------------------------------------------
// qmul : Multiplication de quaternion
//-------------------------------------------------------------------
// sa = sb*sc - Vb'*Vc et Va = sa*Vb + sb*Va + Vb ^ Vc
// Qa(1,1)   = Qb(1)*Qc(1) - Qb(2:4)'*Qc(2:3);
// Qa(2:4,1) = Qb(1)*Qc(2:4) + Qc(1)*Qb(2:4) + cross(Qb(2:4),Qc(2:4));
function Qa = qmul(Qb,Qc)
	Qa = Qc(1)*Qb ; // Permet de dimensionner Qa
	Qa(1) = Qa(1) - Qb(2)*Qc(2) - Qb(3)*Qc(3) - Qb(4)*Qc(4) ;
	Qa(2) = Qa(2) + Qb(1)*Qc(2) + Qb(3)*Qc(4) - Qb(4)*Qc(3) ;
	Qa(3) = Qa(3) + Qb(1)*Qc(3) + Qb(4)*Qc(2) - Qb(2)*Qc(4) ;
	Qa(4) = Qa(4) + Qb(1)*Qc(4) + Qb(2)*Qc(3) - Qb(3)*Qc(2) ;
endfunction;

//--------------------------------------
// Quaternion issu d'un vecteur rotation
//--------------------------------------
function q = vrot2qrot(w)
    q = zeros(4,1);
    ang = norm(w);
    q(1) = cos(ang/2.);
    if ang > 0. then q(2:4) = (sin(ang/2.)/ang)*w(1:3); end
endfunction

//------------------------------------------
// Quaternion issu d'une matrice de rotation
//------------------------------------------
function q = mrot2qrot(a)
// File name:       geouti.sci
//
// Function:        mrot2qrot
//
// Description:     quaternion(4x1) associé à une matrice de rotation(3x3).
//
// Annotations:     Ce code écrit en Scilab a un but purement pédagogique
//
// References:      http://michel.llibre.pagesperso-orange.fr/docs/quaternions.pdf
//
// Author:          Michel Llibre
//
// Date:            Juillet 2011
// 
    q = zeros(4,1) ;
    q(2) =  a(1,1) - a(2,2) - a(3,3) ;
    q(3) = -a(1,1) + a(2,2) - a(3,3) ;
    q(4) = -a(1,1) - a(2,2) + a(3,3) ;
    q(1) =  a(1,1) + a(2,2) + a(3,3) ;

    i = 1 ; for j = 2:4, if q(j) > q(i) then i = j; end, end

    if i == 1 then // q0 est le plus grand 
        q(1) = sqrt(q(1)+1.)/2. ;
        us4qi = 0.25/q(1) ;
        q(2) = us4qi*( a(3,2) - a(2,3) ) ;
        q(3) = us4qi*( a(1,3) - a(3,1) ) ;
        q(4) = us4qi*( a(2,1) - a(1,2) ) ;

    elseif i == 2 then //  q1 est le plus grand 
        q(2) = sqrt(q(2)+1.)/2. ;
        us4qi = 0.25/q(2) ;
        q(1) = us4qi*( a(3,2) - a(2,3)) ;
        q(3) = us4qi*( a(2,1) + a(1,2)) ;
        q(4) = us4qi*( a(1,3) + a(3,1)) ;

    elseif i == 3 then //  q2 est le plus grand 
        q(3) = sqrt(q(3)+1.)/2. ;
        us4qi = 0.25/q(3) ;
        q(1) = us4qi*( a(1,3) - a(3,1)) ;
        q(4) = us4qi*( a(3,2) + a(2,3)) ;
        q(2) = us4qi*( a(2,1) + a(1,2)) ;

    else // q3 est le plus grand */
        q(4) = sqrt(q(4)+1.)/2. ;
        us4qi = 0.25/q(4) ;
        q(1) = us4qi*( a(2,1) - a(1,2)) ;
        q(2) = us4qi*( a(1,3) + a(3,1)) ;
		q(3) = us4qi*( a(3,2) + a(2,3)) ; // ligne oubliée dans la version précédente. 
        //Merci à F. Mignard qui m'a signalé cet oubli.
    end
endfunction

//---------------------------------------------------------
// Quaternion de rotation d'angle a autour d'un axe de base
// L'axe de base (argument i) peut être soit :
// un entier : 1, 2 ou 3
// une des lettres suivantes : 'x', 'y','z', 'i', 'j', ou 'k' 
// exemples : 
// angix2qrot(%pi/3, 1) : Rotation 60 degrés autour de Ox
// angix2qrot(%pi/4, 'y') : Rotation 45 degrés autour de Oy
// angix2qrot(%pi/6, 'k') : Rotation 60 degrés autour de Oz
//-----------------------------------------------------------
function q = angix2qrot(a,i)
    nok = %t
    if type(i) == 10 then
        if     i == 'i' | i == 'I' | i == 'x' | i == 'X' then
            n = 1; nok = %f;
        elseif i == 'j' | i == 'J' | i == 'y' | i == 'Y' then
            n = 2; nok = %f;
        elseif i == 'k' | i == 'K' | i == 'z' | i == 'Z' then
            n = 3; nok = %f;
        end
    elseif type(i) == 1 & i >= 1 & i <= 3 then
        n = i; nok = %f; 
    end
    if nok then
        error("argument axe incorrect !")
    end
    q = zeros(4,1);
    q(1) = cos(a/2.);
    q(n+1) = sin(a/2.);
endfunction

//-----------------------------------------------------------
// Quaternion de rotation d'angle a autour d'un vecteur unité
//-----------------------------------------------------------
function q = angvu2qrot(a, vu)
    q = zeros(4,1);
    q(1) = cos(a/2.);
    q(2:4) = sin(a/2.)*matrix(vu,3,1);
endfunction

//-------------------------------------------------
// Constructeur général d'un quaternion de rotation
//-------------------------------------------------
function q = qrot(a1, a2)
    %nargin = argn(2);
    nok = %t;
    if %nargin == 1 // cas qrot(V3) ou qrot(m33)
        n = length(a1);
        if n == 3 then // qrot(V3)
            nok = %f; q = vrot2qrot(a1);
        elseif n == 9 then // qrot(m33)
            nok = %f; q = mrot2qrot(a1);
        end
    elseif %nargin == 2
        n = length(a2);
        if  n == 1 then // Rotation angle a1 autour de l'axe numero a2
            nok = %f; q = angix2qrot(a1, a2);
        elseif n == 3 then // Rotation angle a1 autour du vecteur a2
            nok = %f; q = angvu2qrot(a1, a2/norm(a2));
        end
    elseif %nargin == 0
        // sans argument -> quaternion identité par défaut
        nok = %f; q = [1;0;0;0];	
    end
    if nok then
        error("Argument(s) création qrot incorrect(s) !");
    end
endfunction

//----------------------------------------
// Matrice antisymétrique associée à un v3
//             [ 0 -Vz Vy]
//         M = [ Vz 0 -Vx]
//             [-Vy Vx 0 ]
//----------------------------------------
function m = manti(v)
    m = [0 -v(3) v(2) ; v(3) 0 -v(1) ; -v(2) v(1) 0] ;
endfunction

//-----------------------------------------------------
// Matrice de rotation à partir q'un vrot
//-----------------------------------------------------
function m = vrot2mrot(v)
    a = norm(v) ; U = manti(v/ang) ; 
    m = eye(3,3) + sin(a)*U + (1.-cos(a))*U*U ;
endfunction

//-----------------------------------------------------------
// Quaternion de rotation d'angle a autour d'un vecteur unité
//-----------------------------------------------------------
function m = angvu2mrot(a, u)
      U = manti(u) ; 
      m = eye(3,3) + sin(a)*U + (1.-cos(a))*U*U ;    
endfunction

//-----------------------------------------------------
// Matrice de rotation à partir q'un qrot
//-----------------------------------------------------
function m = qrot2mrot(q)
	V = manti(q(2:4));
	m = eye(3,3) + 2*(q(1)*V + V*V) ;
endfunction

//-----------------------------------------------------------
// Matrice de rotation d'angle a autour d'un axe de base
// L'axe de base (argument i) peut être soit :
// un entier : 1, 2 ou 3
// une des lettres suivantes : 'x', 'y','z', 'i', 'j', ou 'k' 
// exemples : 
// ai2mr(%pi/3, 1) : Rotation 60 degrés autour de Ox
// ai2mr(%pi/4, 'y') : Rotation 45 degrés autour de Oy
// ai2mr(%pi/6, 'k') : Rotation 60 degrés autour de Oz
//-----------------------------------------------------------
function m = angix2mrot(a,i)
    nok = %t
    if type(i) == 10 then
        if     i == 'i' | i == 'I' | i == 'x' | i == 'X' then
            n = 1; nok = %f;
        elseif i == 'j' | i == 'J' | i == 'y' | i == 'Y' then
            n = 2; nok = %f;
        elseif i == 'k' | i == 'K' | i == 'z' | i == 'Z' then
            n = 3; nok = %f;
        end
    elseif type(i) == 1 & i >= 1 & i <= 3 then
        n = i; nok = %f; 
    end
    if nok then
        error("argument axe incorrect !")
    end
    c = cos(a) ; s = sin(a) ;
    if n == 1 then
        m = [1 0 0 ; 0 c -s ; 0 s c] ;
    elseif n == 2 then
        m = [c 0 s ; 0 1 0 ; -s 0 c] ;
    else
        m = [c -s 0 ; s c 0 ; 0 0 1] ;
    end
endfunction

//-------------------------------------------------
// Constructeur général d'une matrice de rotation
//-------------------------------------------------
function m = mrot(a1, a2)
// File name:       geouti.sci
//
// Function:        mrot
//
// Description:     construction d'une matrice de rotation
//
// Annotations:     Ce code écrit en Scilab a un but purement pédagogique
//
// References:      http://michel.llibre.pagesperso-orange.fr/docs/quaternions.pdf
//
// Author:          Michel Llibre
//
// Date:            Juillet 2011
// 
    %nargin = argn(2);
    nok = %t;
    if %nargin == 1 // cas mrot(V3) ou mrot(qrot)
        n = length(a1);
        if n == 3 then // mrot(V3)
            nok = %f; m = vrot2mrot(a1);
        elseif n == 4 then // mrot(q)
            nok = %f; m = qrot2mrot(a1);
        end
    elseif %nargin == 2
        n = length(a2);
        if  n == 1 then // Rotation angle a1 autour de l'axe numero a2
            nok = %f; m = angix2mrot(a1, a2);
        elseif n == 3 then // Rotation angle a1 autour du vecteur a2
            nok = %f; m = angvu2mrot(a1, a2/norm(a2));
        end
    elseif %nargin == 0
        // sans argument -> quaternion identité par défaut
        nok = %f; m = eye(3,3);	
    end
    if nok then
        error("Argument(s) création mrot incorrect(s) !");
    end
endfunction

//------------------------------------------------------------------------------
// Extraction des paramètres de rotation d'une matrice ou d'un quaternion
//------------------------------------------------------------------------------
//   V=vrot(M)       Vecteur rotation associe a la matrice unitaire M
//   V=vrot(Q)       Vecteur rotation associe au quaternion unitaire Q
//   [a,U]=vrot(M)   Angle de rotation et vecteur unit. de la matrice unitaire M
//   [a,U]=vrot(Q)   Angle de rotation et vecteur unit. du quaternion unitaire Q
//   Angs=vrot(M,'uvw')  Angles d'Euler-Cardan extraits de la matrice unitaire M
//           les lettres u, v et w specifient les axes successifs de
//           rotation. Elles doivent etre une des trois lettres x, y ou z.
//   Angs=vrot(Q,'uvw')  Angles d'Euler-Cardan extraits du quaternion unitaire Q
//           les lettres u, v et w specifient les axes successifs de
//           rotation. Elles doivent etre une des trois lettres x, y ou z.
//------------------------------------------------------------------------------
function [v, w] = vrot(b,c)
// File name:       geouti.sci
//
// Function:        vrot
//
// Description:     Paramètres d'une rotation
//
// Annotations:     Ce code écrit en Scilab a un but purement pédagogique
//
// References:      http://michel.llibre.pagesperso-orange.fr/docs/quaternions.pdf
//
// Author:          Michel Llibre
//
// Date:            Juillet 2011
//     
    [%nargout, %nargin] = argn(0);
    if %nargin == 1
      [li, co] = size(b) ;
      if li == 3 & co == 3 //Argument matrice rotation
        w = zeros(3,1) ;
        w(1) = b(3,2) - b(2,3) ;
        w(2) = b(1,3) - b(3,1) ;
        w(3) = b(2,1) - b(1,2) ;
        ds = norm(w) ; a = asin(ds/2) ;
        if ds ~= 0, w = w/ds ; else w(1) = 0 ; end
      elseif li*co == 4 //argument quaternion
        w = b(2:4) ;
        sas2 = norm(w) ; a = 2*asin(sas2) ;
        if sas2 ~= 0, w = w/sas2 ;
        else w(1) = 0 ; end
        if b(1) < 0., w = -w; end
      else
        error('Argument incorrect') ;
      end
      if %nargout ~= 2
        v = a*w ;
      else
        v = a ;
      end
    elseif %nargin == 2 & type(c)==10 & length(c) == 3
      v = zeros(3,1) ;
      [li, co] = size(b) ;
      if li*co == 4 //argument quaternion
         b = mrot(b) ;
      elseif li ~= 3 | co ~= 3 //Argument matrice rotation
         error('Premier argument incorect')
      end
      c = convstr(c,"l") ; n = ascii(c)-ascii('www') ;
      i = n(1) ; j = n(2) ; k = n(3) ;
      function r = rem(x,y), r = x - fix(x./y).*y, endfunction
      if i == k //Euler 1er type
        v(2) = acos(b(i,k)) ;
        if k == 1+rem(j,3)
          l = 1+rem(i,3) ;
          v(1) = atan(b(j,k), b(l,k)) ;
          v(3) = atan(b(i,j), -b(i,l)) ;
        else
          l = 1+rem(j,3) ;
          v(1) = atan(b(j,k), -b(l,k)) ;
          v(3) = atan(b(i,j), b(i,l)) ;
        end
      else //Euler 2eme type
        if k == 1+rem(i,3)
          v(1) = atan(b(j,k),b(k,k)) ;
          v(2) = - asin(b(i,k)) ;
          v(3) = atan(b(i,j),b(i,i)) ;
        else
          v(1) = - atan(b(j,k),b(k,k)) ;
          v(2) = asin(b(i,k)) ;
          v(3) = - atan(b(i,j),b(i,i)) ;
        end
      end
    else
      disp(type(c))
      disp(size(c,2))
      error('Arguments incorrects') ;
    end
endfunction

//-----------------------------------------------------------------------
//    Qa = qprox(Qb, Qc)      Qa = (+/-)Qb (le plus proche de Qc)
//    pour prendre la representation d'orientation Qa de l'orientation B
//    definie par Qb la plus proche de la representation Qc d'une certaine
//    orientattion C.
//    En effet kQ (avec k scalaire quelconque et Q quaternion unitaire)
//    represente la meme orientation que Q, ce qui fait une infinite de
//    representations. En se limitant aux representation unitaires, il
//    reste quand meme Q et -Q. 
//-----------------------------------------------------------------------
function [Qa] = qprox(Qb,Qc)
		
s = Qb(1)*Qc(1)+Qb(2)*Qc(2)+Qb(3)*Qc(3)+Qb(4)*Qc(4) ;
if s < 0
   Qa = - Qb ;
else
   Qa = Qb ; 
end
endfunction

//-----------------------------------------------------------------------
// Carol = mat2carol(MatRot, Carol)
//----------------------------------------------------------------------- 
// Extraction des angles de Cap, assiette et Roulis (ordre ZYX)
// de la matrice MatRot
// Cette routine est une version améliorée de vrot(MatRot,'zyx') qui fournit
// toujours la solution de la première nappe (avec l'assiette inférieure à pi/2
// en valeur absolue) et ne traite pas la congiguration singulière (assiette
// égale à +/- pi/2)
// Dans cette version : 
// - Pour pi/2 - abs(assiette) < 1.E-5  on cherche la continuité avec les
// valeurs fournies en argument (ou 0,0,0 si absentes) en conservant la somme 
// ou cap et roulis.
// Les valeurs passées en référence sont également utilisées pour envoyer la
// solution dans la même nappe que ces valeurs
//-----------------------------------------------------------------------
function Carol = mat2carol(a,Carol)
// File name:       geouti.sci
//
// Function:        mat2carol
//
// Description:     Cap, assiette roulis d'une rotation
//
// Annotations:     Ce code écrit en Scilab a un but purement pédagogique
//
// References:      http://michel.llibre.pagesperso-orange.fr/docs/quaternions.pdf
//
// Author:          Michel Llibre
//
// Date:            Juillet 2011
//     
    
    %nargin = argn(2);
    if %nargin == 1
       Carol = zeros(3,1);
    end
    
    tet = -asin(a(3,1));
    if abs(tet) < %pi/2-1.e-7
       psi = atan(a(2,1),a(1,1));
       phi = atan(a(3,2),a(3,3));
    else
       if tet > 0.
          som = Carol(1)+Carol(3);
          dif = atan(-a(1,2),a(2,2));
       else
          dif = Carol(1)-Carol(3);
          som = atan(-a(1,2),a(2,2));
       end
       psi = (som+dif)/2.;
       phi = (som-dif)/2.;
    end
    // On met dans la même nappe que la solution précédente
    if cos(Carol(2)) < 0.
       psi = psi+%pi;
       phi = phi+%pi;
       tet = %pi-tet;
    end
    // On démodule 2*pi
    while (psi-Carol(1)) >  %pi, psi = psi - 2*%pi; end
    while (psi-Carol(1)) < -%pi, psi = psi + 2*%pi; end
    while (tet-Carol(2)) >  %pi, tet = tet - 2*%pi; end
    while (tet-Carol(2)) < -%pi, tet = tet + 2*%pi; end
    while (phi-Carol(3)) >  %pi, phi = phi - 2*%pi; end
    while (phi-Carol(3)) < -%pi, phi = phi + 2*%pi; end
    Carol(1) = psi;
    Carol(2) = tet;
    Carol(3) = phi;
endfunction
    