//==============================================================================
// Utilitaires pour manipuler les quaternions comme des nombres 
// Copyright (C) Michel Llibre.
//==============================================================================
// Dans ce fichier, les quaternions sont programmés au moyen de "mlist" qui
// permettent la surcharge des opérateurs. On a ainsi la possibilité de combiner
// les quaternions entre eux ou avec des scalaires au moyen des 4 opérations
// comme des nombres ordinaires.
// Dans la mlist q le quaternion est stocké sous forme d'un vecteur colonne 4x1
// qui est directement accessible par q.h ou par la fonction q2R4(q)
//
// Cette programmation effectuée par un béotien en Scilab, n'est pas du tout
// optimisée. Elle a un but purement pédagogique. Par ailleurs, il manque
// certaines vérifications sur les types et les dimensions qui risquent de
// donner des résultats aberrants en cas d'utilisation avec des arguments
// incorrects.
//
// ATTENTION ATTENTION ATTENTION ATTENTION ATTENTION ATTENTION ATTENTION
// Les fonctions du fichier geouti.sci ne savent pas traiter ces quaternions.
// Elles traitent de simples vecteurs 4x1 qui sont la partie numérique du
// quaternion de type mlist défini dans ce fichier-ci. Répétons :
// - Les quaternions `qr' renvoyés par les fonctions définies dans le
// fichier geouti.sci sont des vecteurs 4x1. Pour les convertir en quaternion
// `q' de type mlist utilisables comme des nombres, faire q = quat(qr).
// - De même, les arguments quaternions `qr' utisés par les fonctions du fichier
// geouti.sci sont également des vecteurs 4x1. Pour leur fournir ce type
// d'argument à partir d'un quaternion `q' de type mlist, fournir q2R4(q) ou
// ce qui revient au même q.h.
//==============================================================================
// Les fonctions fournies sont les suivantes :
// Construction d'un quaternion :
// - a partir de ses composantes : 
//        q = quat();            // quaterrnion identité
//        q = quat(s,x,y,z);
//        q = quat([s,x,y,z]);
//        q = quat([s;x;y;z]);
//        q = quat(x,y,z);        // quaternion pur
//        q = quat([x,y,z]);      // quaternion pur
//        q = quat([x;y;z]);      // quaternion pur
// - à partir d'une matrice de rotation :
//        q = quat(mrot);
// - à partir d'un angle et du numéro de l'axe de base :
//        q = quat(%pi/4, 2);     // rotation de 45 degrés autour de Oy
// - à partir d'un angle et du vecteur axial :
//        q = quat(%pi/6, [1,0,1]) // rotation de 30 degrés autour de la 1ère 
//             bissectrice du plan xOz.
// - à partir d'un vecteur rotation :
//        q = vr2quat(V); // V : direction de l'axe et angle = norm(V)
// Display :
//        disp(q);
// Accès aux composantes :
//        v = q(2:4); // exemple d'accès à la partie pure
// Reconfiguration en vecteur (colonne) ordinaire de R4 :
//        W = q2R4(Q); // ou plus simplement Q.h
// Conjuquaison :
//        q1 = q';
// Inversion :
//        q2 = inv(q);
// Négation, soustraction, addition :
//        q = -q1 + q2 - q3
// Multiplication, division par un scalaire ou un autre quaternion :
//        q = 2*q1*q2/q3/3;
// Unitarisation :
//        q = unit(q);
//==============================================================================

exec("geouti.sci",-1); // Utilitaires avec des quaternions simples vecteurs 4x1.

//---------------------------------------------------------------
// Prototype de la définition d'un quaternion unité par une mlist
//---------------------------------------------------------------
Q = mlist(['quat','h'],[1,0,0,0]);

//---------------------------------
// Surcharge de la fonction display
//---------------------------------
function [] = %quat_p(Q)
    disp([Q.h']);
endfunction

//--------------------------------------------------------------
// Fonction construisant des quaternions avec de 0 à 4 arguments 
//--------------------------------------------------------------
function Q = quat(a1,a2,a3,a4)
// File name:       quatern.sci
//
// Function:        quat
//
// Description:     construction d'un quaternion (0 à 4 arguments)
//
// Annotations:     Ce code écrit en Scilab a un but purement pédagogique
//
// References:      http://michel.llibre.pagesperso-orange.fr/docs/quaternions.pdf
//
// Author:          Michel Llibre
//
// Date:            Juillet 2011
//    
    [%nargout, %nargin] = argn(0);
    nok = %t;
    
    if %nargin == 1
        n = length(a1);
        if n == 4 then // vecteur 4 composantes -> quaternion
            Q = mlist(['quat','h'], matrix(a1,4,1));
            nok = %f;
        elseif n == 3 then // vecteur 3 composantes -> quaternion pur
            Q = mlist(['quat','h'],[0; matrix(a1,3,1)]);
            nok = %f;
        elseif n == 9 then // mrot -> quat
            Q = mlist(['quat','h'], mrot2qrot(a1));
            nok = %f
        end
    elseif %nargin == 2
        n = length(a2);
        if  n == 1 then // Rotation angle a1 autour de l'axe numero a2
            Q = mlist(['quat','h'], angix2qrot(a1,a2));
            nok = %f;
        elseif n == 3 then // Rotation angle a1 autour du vecteur a2
            u = a2/norm(a2);
            Q = mlist(['quat','h'], angvu2qrot(a1,u));
            nok = %f;
        end
    elseif %nargin == 3 // 3 composantes -> quaternion pur
        Q = mlist(['quat','h'],[0; a1; a2; a3]);
        nok = %f;	
    elseif %nargin == 4 // 4 composantes -> quaternion
        Q = mlist(['quat','h'],[a1; a2; a3; a4]);
        nok = %f;
    elseif %nargin == 0
        // sans argument -> quaternion identité par défaut
        Q = mlist(['quat','h'],[1;0;0;0]);
        nok = %f;	
    end
    if nok then
        error("Argument(s) création quat incorrect(s) !");
    end
endfunction

// Pour le construction d'un quaternion à partir d'un v3 nous
// avons choisi de générer le quaternion pur associé au vecteur
// plutot que le quaternion de rotation associé au vecteur rotation.
// pour générer celui-ci, nous fournissons la fonction suivante.
//---------------------------------------------------
// Creation quaternion à partir d'un vecteur rotation
//---------------------------------------------------
function Q = vr2quat(V)
    Q = mlist(['quat','h'], vrot2qrot(V));
endfunction

//----------------------------------
// Conversion en vecteur dimension 4
//----------------------------------
function q = q2R4(Q)
    q = Q.h;
endfunction 

//----------------------------------
// Extraction composante (surcharge)
//----------------------------------
function r = %quat_e(i,Q)
    // pour les adeptes de la sécurité décommenter les 3 lignes suivantes
    //    if i <= 0 | i > 4 then
    //        error("Indice hors domaine !")
    //    end
    r = Q.h(i);
endfunction

//-----------------------------
// Soustraction : surcharge Q-R
//-----------------------------
function Q = %quat_s_quat(Q,R)
    Q = mlist(['quat','h'], Q.h - R.h);
endfunction

//-------------------------
// Addition : surcharge Q+R
//-------------------------
function Q = %quat_a_quat(Q,R)
    Q = mlist(['quat','h'], Q.h + R.h);
endfunction

//------------------------
// Négation : surcharge -R
//------------------------
function Q = %quat_s(R)
    Q = mlist(['quat','h'], -R.h);
endfunction

//-------------------------------
// Multiplication : surcharge Q*R
//-------------------------------
function Q = %quat_m_quat(Q,R)

//    s1 = Q.h(1); x1 = Q.h(2); y1 = Q.h(3); z1 = Q.h(4);
//    s2 = R.h(1); x2 = R.h(2); y2 = R.h(3); z2 = R.h(4);
//
//    s = s1*s2 - x1*x2 - y1*y2 - z1*z2 ;
//    x = s1*x2 + s2*x1 + y1*z2 - z1*y2 ;
//    y = s1*y2 + s2*y1 + z1*x2 - x1*z2 ;
//    z = s1*z2 + s2*z1 + x1*y2 - y1*x2 ;
//    Q = mlist(['quat','h'],[s; x; y; z]);
    Q = mlist(['quat','h'], qmul(Q.h, R.h));
endfunction

//-----------------------------------------
// Conjuguaison : surcharge R' (apostrophe)
//-----------------------------------------
function Q = %quat_t(R)
    Q = mlist(['quat','h'],[R.h(1); -R.h(2:4)]);
endfunction

//-----------------------------
// Inversion : surcharge inv(R)
//-----------------------------
function Q = %quat_inv(R)
    r2 = R.h'*R.h;
    // Attention : pas de test division par zéro.
    Q = mlist(['quat','h'],[R.h(1)/r2; -R.h(2:4)/r2]);
endfunction

//--------------------
// Quaternion unitaire
//--------------------
function Q = unit(R)
    r = norm(R.h);
    if r == 0 then

        Q = mlist(['quat','h'], [1;0;0;0]);
    else
        Q = mlist(['quat','h'], R.h/r);
    end
endfunction

//-------------------------
// Division : surcharge Q/R
//-------------------------
function Q = %quat_r_quat(Q,R)
    Q = Q*inv(R);
endfunction

//----------------------------------------------------------
// Multiplication par un scalaire ou vecteur : surcharge s*R
//----------------------------------------------------------
function Q = %s_m_quat(s,R)
    n = length(s)
    if n == 1 then
        Q = mlist(['quat','h'], s*R.h);
    elseif n == 3 | n == 4 then
        Q = quat(s)*R;
    else
        error("argument multiplication incorrect!");
    end
endfunction

//----------------------------------------------------------
// Multiplication par un scalaire ou vecteur : surcharge R*s
//----------------------------------------------------------
function Q = %quat_m_s(R,s)
    n = length(s)
    if n == 1 then
        Q = mlist(['quat','h'], s*R.h);
    elseif n == 3 | n == 4 then
        Q = R*quat(s);
    else
        error("argument multiplication incorrect!");
    end
endfunction

//----------------------------------------------------
// Division par un scalaire ou vecteur : surcharge R/s
//----------------------------------------------------
function Q = %quat_r_s(R,s)
    n = length(s)
    if n == 1 then
        Q = mlist(['quat','h'], R.h/s);
    elseif n == 3 | n == 4 then
        Q = R*inv(quat(s));
    else
        error("argument division incorrect!");
    end
endfunction

//------------------------------------------------------------
// Division scalaire ou vecteur par quaternion : surcharge s/R
//------------------------------------------------------------
function Q = %s_r_quat(s,R)
    n = length(s)
    if n == 1 | n == 3 | n== 4 then
        Q = s*inv(R)
    else
        error("argument division incorrect!");
    end
endfunction

